%% Applying MPO onto MPS
% Author: <https://www.theorie.physik.uni-muenchen.de/lsvondelft/members/sci_mem/seung_sup-lee/index.html 
% Seung-Sup Lee>
%% 
% We demonstrate that the application of matrix product operator (MPO) onto 
% a matrix product state (MPS) results in another MPS. In this tutorial, we consider 
% an MPO representation of the Hamiltonian and its ground state(s) as the MPO 
% and the MPS, respectively.
% 
% To start with, we consider the spin-1/2 Heisenberg model on a short chain,
% 
% $$H = \sum_{\ell = 1}^{N-1} \hat{\vec{S}}_\ell \cdot \hat{\vec{S}}_{\ell + 
% 1}.$$
% 
% The chain length $N=6$ is chosen to be short so that we can obtain the full 
% Hamiltonian and the exact ground state by using iterative diagonalization. Here 
% at all iterations before the last iteration, we do not rotate the basis into 
% the energy eigenbasis and do not truncate the Hilbert space. It is necessary 
% to compare the Hamiltonian constructed within iterative diagonalization (represented 
% by a matrix |Hnow| below) with the MPO form of the Hamiltonian (to be obtained 
% in the next section).

clear

N = 6; % chain length
M = cell(1,N); % MPS

[S,I] = getLocalSpace('Spin',1/2);

H0 = I*0; % Hamiltonian for only the 1st site
M{1} = getIdentity(1,2,I,2); % 1st leg is dummy leg (vacuum)

for itN = (1:N)
    if itN == 1
        Hnow = H0;
        Anow = M{1};
    else
        % % add new site
        Anow = getIdentity(Hprev,2,I,2);
        Hnow = updateLeft(Hprev,2,Anow,[],[],Anow);
        % update the Hamiltonian up to the last sites
        % to the enlarged Hilbert space
        
        % % spin-spin interaction
        % Hermitian conjugate of the spin operator at
        % the current site
        Sn = permute(conj(S),[3 2 1]);
        HSS = updateLeft(Sprev,3,Anow,Sn,3,Anow);
        Hnow = Hnow+HSS;
    end
    
    if itN < N % no rotation, no truncation
        M{itN} = Anow;
        Hprev = Hnow;
        
        % spin operator at the current site; to be used for 
        % generating the coupling term at the next iteration
        Sprev = updateLeft([],[],M{itN},S,3,M{itN});
    else
        [V,D] = eig((Hnow+Hnow')/2);
        [E_GS,minid] = min(diag(D));
        % select only the ground state
        M{itN} = contract(Anow,3,3,V(:,minid),2,1);
    end
    
    disptime(['#',sprintf('%02i/%02i',[itN,N]),' : ', ...
        'NK=',sprintf('%i/%i',[size(M{itN},3),size(Hnow,2)])]);
end
%% 
% The ground-state energy is:

E_GS
%% 
% Currently, |M| is in a left-canonical form. By bringing it into a right-canonical 
% form, the size of the constituent tensors can be decreased.

M
M = canonForm(M,0) % right-canonical form
%% MPO representation of Heisenberg chain Hamiltonian
% We construct the MPO representation of the chain Hamiltonian, following the 
% recipe given in a lecture. The MPO consists of the same tensor, called bulk 
% tensor, except for the first and the last sites. A bulk tensor is rank-4, and 
% its legs are as left-bottom-right-top. The bottom (top) leg is to be contracted 
% with the second leg of bra (ket) tensor.
% 
% Let's first generate a bulk tensor. 

% bulk tensor for each chain site
Hloc = cell(5,5);
Hloc(:) = {zeros(size(I))};
Hloc{1,1} = I;
for ito = (1:size(S,2)) % different components of spin operators
    Hloc{ito+1,1} = squeeze(S(:,ito,:));
    Hloc{end,ito+1} = squeeze(S(:,ito,:))';
end
Hloc{end,end} = I;
Hloc = cell2mat(reshape(Hloc,[1 1 size(Hloc,1) size(Hloc,2)]));
Hloc = permute(Hloc,[3 1 4 2]); % leg order: left-bottom-right-top
%% 
% For the tensors at the first and the last sites, we project the bulk tensor 
% onto a specific index of its left and right legs, respectively. So the left 
% and right legs, respectively, become dummy legs with singleton dimension.

% MPO for the full chain
Hs = cell(1,N);
Hs(:) = {Hloc};
Hs{1} = Hs{1}(end,:,:,:); % choose the last index of the left leg
Hs{end} = Hs{end}(:,:,1,:); % choose the first index of the right leg
%% 
% The indices are chosen such that the pair of spin operators (which make up 
% the interaction term) are complete, without unpaired operators remaining.
% 
% To check whether the MPO construction is right, we compare the MPO and the 
% Hamiltonian from iterative diagonalization (represented by |Hnow|) We first 
% contract the tensors of |Hs| to make a high-rank tensor acting onto all the 
% chain sites. And we permute and reshape the high-rank tensor into a big matrix, 
% which can be directly compared with |Hnow|.

Hs_tot = permute(Hs{1},[1 2 4 3]); % leg order: left-bottom-top-right
for itN = (2:N)
    Hs_tot = contract(Hs_tot,2*itN,2*itN,permute(Hs{itN},[1 2 4 3]),4,1);
end
Hs_tot = permute(Hs_tot,[(2:2:2*N) (3:2:2*N+1) 1 2*N+2]);
% permute singleton dimensions (= dummy legs) to the end
Hs_tot = reshape(Hs_tot,(size(I,1)^N)*[1 1]);
%% 
% The two forms of the Hamiltonian are equivalent.

max(abs(Hs_tot(:)-Hnow(:)))
%% Application of MPO onto MPS
% We apply the MPO representation of Hamiltonian $H$ (represented by a cell 
% array |Hs|) onto the ground state $|\psi \rangle$ (by a cell array |M|). The 
% tensors at each site, |Hs{n}| and |M{n}|, are contracted, and their horizontal 
% legs are merged by using isometries. The result is a rank-3 tensor |HM{n}|, 
% which constitutes a MPS form of $H |\psi\rangle$.
% 
% 

HM = cell(1,N);
for itN = (1:N)
    HM{itN} = contract(Hs{itN},4,4,M{itN},3,2);
    % leg order: Hleft-Hbottom-Hright-Mleft-Mright
    
    % isometry to merge left legs
    if itN == 1
        Aleft = getIdentity(HM{itN},1,HM{itN},4);
    else
        % use Aright from the previous iteration, 
        % to be a valid insertion of identity
        Aleft = conj(Aright);
    end
    
    % isometry to merge right legs
    Aright = getIdentity(HM{itN},3,HM{itN},5);
    
    % contract isometries
    HM{itN} = contract(Aleft,3,[1 2],HM{itN},5,[1 4]);
    % leg order: left-Hbottom-Hright-Mright
    HM{itN} = contract(HM{itN},4,[3 4],Aright,3,[1 2]);
end
%% 
% As shown in the above tensor diagram, the bond dimensions have increased.

HM
%% 
% The bond dimensions in the MPS form of $H|\psi\rangle$ can be compressed by 
% performing a "round trip" of bringing into canonical forms; first into left-canonical, 
% then into right-canonical.

[HM,HMnorm] = canonForm(HM,N) % left-canonical
HM = canonForm(HM,0) % here the second output should be 1
%% 
% Now, the tensors have the same sizes as those in |M|.

M
%% 
% The value of the norm |HMnorm| (which means $\| H | \psi \rangle \|$) is equal 
% to the absolute value of the ground-state energy (represented by |E_GS|).

HMnorm - abs(E_GS) % zero up to numerical noise
%% 
% For later purpose, we let the first tensor |HM{1}| absorb |HMnorm|.

HM{1} = HM{1}*HMnorm;
%% 
% We can also check that $\langle \psi | H | \psi \rangle$ gives the ground-state 
% energy, by contracting |HM| and |M|.

MHM = updateLeft([],[],M{1},[],[],HM{1});
for itN = (2:N)
    MHM = updateLeft(MHM,2,M{itN},[],[],HM{itN});
end
MHM - E_GS % zero up to numerical noise
%% 
% Here we know that |M| represents the ground state, since the Hamiltonian can 
% be exactly diagonalized, thank to small system size. For general systems, however, 
% there is no straightforward way to verify whether a given state $|\psi\rangle$ 
% is the true ground state of the system, not being a local mimima.
% 
% Still, we can verify whether $|\psi \rangle$ is an eigenstate of the Hamiltonian 
% $H$ by confiming $\langle \psi | H | \psi \rangle^2 - \langle \psi | H^2 | \psi 
% \rangle = 0$. The first term $\langle \psi | H | \psi \rangle^2$ can be computed 
% as the square of |MHM| and the second term $\langle \psi | H^2 | \psi \rangle$ 
% as the squared norm $\| H | \psi \rangle \|^2$. We see that the equality is 
% satisfied, meaning that $|\psi\rangle$ is an eigenstate.

MHM^2 - HMnorm^2
%% Exercise (a): MPO representation of the AKLT Hamiltonian
% We have studied the AKLT states in the previous tutorials. The AKLT states 
% are the ground states of the AKLT model, which is a chain of spin-1's that interact 
% via nearest-neighbor interactions,
% 
% $$\hat{H} = \sum_{\ell=1}^{N-1} \left[ (\hat{\vec{S}}_\ell \cdot \hat\vec{S}_{\ell+1}) 
% + \frac{1}{3} (\hat\vec{S}_\ell \cdot \hat\vec{S}_{\ell+1})^2 \right].$$
% 
% The first term on the right-hand side is the Heisenberg interaction. The second 
% term is a biquadratic term that has a form of the squared Heisenberg interaction. 
% We consider an open boundary condition.
% 
% (i) Construct a bulk tensor of an MPO representation of the AKLT model Hamiltonian. 
% The tensors of the MPO at the first and the last sites can be obtained in the 
% same way as above, by projecting onto specific indices.
% 
% (ii) Check whether your construction of the MPO is correct, by comparing with 
% the Hamiltonian from iterative diagonalization of a short chain, e.g., $N=4$.
%% Exercise (b): Confirm the AKLT states are the eigenstates of the AKLT Hamiltonian
% In a previous tutorial, we have learned that there are four AKLT states $|\psi(\alpha,\beta)\rangle$ 
% ($\alpha,\beta = 1,2$) in case of the open boundary condition. For *all* these 
% AKLT states $|\psi (\alpha,\beta)\rangle$, (i) compute the expectation values 
% $\langle \psi(\alpha,\beta) | H | \psi(\alpha,\beta)\rangle$ and (ii) confirm 
% $\langle \psi | H | \psi \rangle^2 - \langle \psi | H^2 | \psi \rangle = 0$. 
% For this exercise, consider a long chain of length $N = 50$.